/*___INFO__MARK_BEGIN__*/
/*************************************************************************
 *
 *  The Contents of this file are made available subject to the terms of
 *  the Sun Industry Standards Source License Version 1.2
 *
 *  Sun Microsystems Inc., March, 2001
 *
 *
 *  Sun Industry Standards Source License Version 1.2
 *  =================================================
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.2 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://gridengine.sunsource.net/Gridengine_SISSL_license.html
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *   The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *   Copyright: 2001 by Sun Microsystems, Inc.
 *
 *   All Rights Reserved.
 *
 ************************************************************************/
/*___INFO__MARK_END__*/                                   

#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>


#include "sge_bootstrap.h"
#include "sge_unistd.h"

#include "sgermon.h"
#include "sge_log.h"

#include "sge_spool.h"
#include "sge_cqueue_qmaster.h"

#include "cull.h"

#include "commlib.h"
#include "commd_message_flags.h"

#include "sge_host.h"

#include "sge.h"
#include "spool/sge_dirent.h"
#include "sort_hosts.h"          /* JG: TODO: from libsched. Do we need it for spooling? */
#include "sge_complex_schedd.h"  /* JG: TODO: dito */
#include "sge_select_queue.h"    /* JG: TODO: dito */

#include "sge_answer.h"
#include "sge_calendar.h"
#include "sge_ckpt.h"
#include "sge_centry.h"
#include "sge_conf.h"
#include "sge_pe.h"
#include "sge_cqueue.h"
#include "sge_qinstance.h"
#include "sge_qinstance_state.h"
#include "sge_userprj.h"
#include "sge_userset.h"
#include "sge_utility.h"
#include "sge_resource_quota.h"

#include "sge_str.h"

#include "sge_queue_event_master.h"

#include "read_write_cal.h"
#include "read_write_ckpt.h"
#include "read_write_complex.h"
#include "rw_configuration.h"
#include "read_write_host.h"
#include "read_write_pe.h"
#include "read_write_userprj.h"
#include "read_write_userset.h"
#include "read_write_cqueue.h"
#include "read_write_centry.h"
#include "sge_cuser.h"
#include "read_write_ume.h"

#include "sge_hgroup.h"
#include "read_write_host_group.h"
#include "read_write_qinstance.h"
#include "read_write_resource_quota.h"

#include "setup_path.h"
#include "sge_uidgid.h"

#include "msg_common.h"
#include "msg_spoollib_classic.h"

#include "read_list.h"

#define CONFIG_TAG_OBSOLETE_VALUE         0x0001

static int reresolve_host(lListElem *ep, int nm, const char *object_name, const char *object_dir);


int sge_read_host_group_entries_from_disk()
{ 
  lList*     direntries = NULL; 
  lListElem* direntry = NULL;
  lListElem* ep = NULL;
  const char*      hostGroupEntry = NULL;
  int        ret = 0;  /* 0 means ok */
  lList **master_hgroup_list = object_type_get_master_list(SGE_TYPE_HGROUP);

  DENTER(TOP_LAYER, "sge_read_host_group_entries_from_disk");
 
 
  direntries = sge_get_dirents(HGROUP_DIR);
  if (direntries) {
     if (*master_hgroup_list == NULL) {
        *master_hgroup_list = lCreateList("", HGRP_Type);
     }  
     if (!sge_silent_get()) { 
        printf("%s\n", MSG_CONFIG_READINGHOSTGROUPENTRYS);
     }
     
     for_each(direntry, direntries) {
        hostGroupEntry = lGetString(direntry, ST_name);

        if (hostGroupEntry[0] != '.') {
           if (!sge_silent_get()) { 
              printf("\t");
              printf(MSG_SETUP_HOSTGROUPENTRIES_S, hostGroupEntry);
              printf("\n");
           }

           ep = cull_read_in_host_group(HGROUP_DIR, hostGroupEntry , 1, 0, NULL, NULL); 
           if (strcmp(hostGroupEntry, lGetHost(ep, HGRP_name))) {
               ERROR((SGE_EVENT, MSG_HGROUP_INCFILE_S, hostGroupEntry));
               DRETURN(-1);
           }
           lAppendElem(*master_hgroup_list, ep);
        } else {
           sge_unlink(HGROUP_DIR, hostGroupEntry);
        }   
     } 
     lFreeList(&direntries);
  }

  /* everything is done very well ! */
  DRETURN(ret);
}

#ifndef __SGE_NO_USERMAPPING__

int sge_read_user_mapping_entries_from_disk()
{ 
  lList*     direntries = NULL; 
  lListElem* direntry = NULL;
  lListElem* ep = NULL;
  const char*      ume = NULL;
  int        ret = 0;  /* 0 means ok */

  DENTER(TOP_LAYER, "sge_read_user_mapping_entries_from_disk");
 
  direntries = sge_get_dirents(UME_DIR);
  if (direntries) {
     lList **cuser_master_list = object_type_get_master_list(SGE_TYPE_CUSER); 
     if (*(cuser_master_list) == NULL) {
        *(cuser_master_list) = lCreateList("", CU_Type);
     }  
     if (!sge_silent_get()) { 
        printf("%s\n", MSG_CONFIG_READINGUSERMAPPINGENTRY);
     }
     
     for_each(direntry, direntries) {
         ume = lGetString(direntry, ST_name);

         if (ume[0] != '.') {
            if (!sge_silent_get()) { 
               printf("\t");
               printf(MSG_SETUP_MAPPINGETRIES_S, ume);
               printf("\n");
            }

            ep = cull_read_in_ume(UME_DIR, ume , 1, 0, NULL, NULL); 
         
            if (ep != NULL) {
               lAppendElem(*(cuser_master_list), ep);
            } else {
               WARNING((SGE_EVENT, MSG_ANSWER_IGNORINGMAPPINGFOR_S,  ume ));  
               lFreeElem(&ep);
               ep = NULL; 
            } 
         } else {
            sge_unlink(UME_DIR, ume);
         }
     } 
     lFreeList(&direntries);
  }
  
  /* everything is done very well ! */
  DRETURN(ret);
}

#endif

int sge_read_exechost_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;
   const char *host;

   DENTER(TOP_LAYER, "sge_read_exechost_list_from_disk");

   /* 
   ** read exechosts into Master_Exechost_List 
   */
   if (*list == NULL) {
      *list = lCreateList("", EH_Type);
   }

   direntries = sge_get_dirents(directory);
   if(direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINEXECUTIONHOSTS);
      
      for_each(direntry, direntries) {

         host = lGetString(direntry, ST_name);
         if (host[0] != '.') {
            lList *master_centry_list = *object_type_get_master_list(SGE_TYPE_CENTRY);
            
            DPRINTF(("Host: %s\n", host));

            ep = cull_read_in_host(directory, host, CULL_READ_SPOOL, EH_name, 
                                   NULL, NULL);
            if (!ep) {
               lFreeList(&direntries);
               DRETURN(-1);
            }

           /* resolve hostname anew */
            if (reresolve_host(ep, EH_name, "exec host", directory)) {
               lFreeElem(&ep);
               lFreeList(&direntries);
               DRETURN(-1); /* general problems */
            }

            /* necessary to setup actual list of exechost */
            debit_host_consumable(NULL, ep, master_centry_list, 0);

            /* necessary to init double values of consumable configuration */
            centry_list_fill_request(lGetList(ep, EH_consumable_config_list), 
                                     NULL, master_centry_list, true, false, true);

            if (ensure_attrib_available(NULL, ep, EH_consumable_config_list)) {
               lFreeElem(&ep);
               lFreeList(&direntries);
               DRETURN(-1);
            }

            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, host);
         }
      }

      /* merge global and local objects */
      host_list_merge(*list);
      
      lFreeList(&direntries);
   }

   DRETURN(0);
}

int sge_read_adminhost_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;
   const char *host;

   DENTER(TOP_LAYER, "sge_read_adminhost_list_from_disk");

   /* 
   ** read adminhosts into Master_Adminhost_List 
   */
   if (*list == NULL) {
      *list = lCreateList("", AH_Type);
   }

   direntries = sge_get_dirents(directory);
   if(direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINADMINHOSTS);
      for_each(direntry, direntries) {
         host = lGetString(direntry, ST_name);

         if (host[0] != '.') {
            DPRINTF(("Host: %s\n", host));
            ep = cull_read_in_host(directory, host, CULL_READ_SPOOL, AH_name, NULL, NULL);
            if (!ep) {
               lFreeList(&direntries);
               DRETURN(-1);
            } 

            /* resolve hostname anew */
            if (reresolve_host(ep, AH_name, "admin host", directory)) {
               lFreeList(&direntries);
               DRETURN(-1); /* general problems */
            }

            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, host);  
         }
      }
      lFreeList(&direntries);
   }

   DRETURN(0);
}

int sge_read_submithost_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;
   const char *host;

   DENTER(TOP_LAYER, "sge_read_submithost_list_from_disk");

   /* 
   ** read submithosts into Master_Submithost_List 
   */
   if (*list == NULL) {
      *list = lCreateList("", SH_Type);
   }

   direntries = sge_get_dirents(directory);
   if(direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINSUBMITHOSTS);
      for_each(direntry, direntries) {
         host = lGetString(direntry, ST_name);
         if (host[0] != '.') {
            DPRINTF(("Host: %s\n", host));
            ep = cull_read_in_host(directory, host, CULL_READ_SPOOL, 
               SH_name, NULL, NULL);
            if (!ep) {
               lFreeList(&direntries);
               DRETURN(-1);
            } 

            /* resolve hostname anew */
            if (reresolve_host(ep, SH_name, "submit host", directory)) {
               DRETURN(-1); /* general problems */
            }

            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, host);             
         }
      }
      lFreeList(&direntries);
   }

   DRETURN(0);
}

int sge_read_pe_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;
   int ret = 0;
   const char *pe;

   DENTER(TOP_LAYER, "sge_read_pe_list_from_disk");
   
   if (*list == NULL) {
      *list = lCreateList("", PE_Type);
   }

   direntries = sge_get_dirents(directory);
   if(direntries) {
      if (!sge_silent_get()) {
         printf("%s\n", MSG_CONFIG_READINGINGPARALLELENV);
      }
      for_each(direntry, direntries) {
         pe = lGetString(direntry, ST_name);
         if (pe[0] != '.') {
            if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_PE_S, pe);
               printf("\n");
            }
            if (verify_str_key(NULL, pe, MAX_VERIFY_STRING,
                               "pe", KEY_TABLE) != STATUS_OK) {
               DRETURN(-1);
            }       
            ep = cull_read_in_pe(directory, pe, 1, 0, NULL, NULL);
            if (!ep) {
               ret = -1;
               break;
            }

            if (pe_validate(ep, NULL, 1)!=STATUS_OK) {
               ret = -1;
               break;
            }
            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, pe);
         }
      }
      lFreeList(&direntries);
   }

   DRETURN(ret);
}



int sge_read_cal_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *alp = NULL;
   lList *direntries;
   lListElem *aep, *ep, *direntry;
   int ret = 0;
   const char *cal;
   const char *s;

   DENTER(TOP_LAYER, "sge_read_cal_list_from_disk");
   
   if (*list == NULL) {
      *list = lCreateList("", CAL_Type);
   }

   direntries = sge_get_dirents(directory);
   if(direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINCALENDARS);
      for_each(direntry, direntries) {
         cal = lGetString(direntry, ST_name);

         if (cal[0] != '.') {
            if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_CALENDAR_S, cal);
               printf("\n");
            }
            if (verify_str_key(NULL, cal, MAX_VERIFY_STRING,
                               "cal", KEY_TABLE) != STATUS_OK) {
               DRETURN(-1);
            }      
            ep = cull_read_in_cal(directory, cal, 1, 0, NULL, NULL);
            if (!ep) {
               ret = -1;
               break;
            }

            if (!calendar_parse_year(ep, &alp) || 
                !calendar_parse_week(ep, &alp)) {
               if (!(aep = lFirst(alp)) || !(s = lGetString(aep, AN_text)))
                  s = MSG_UNKNOWNREASON;
               ERROR((SGE_EVENT,MSG_CONFIG_FAILEDPARSINGYEARENTRYINCALENDAR_SS, 
                     cal, s));
               lFreeList(&alp);
               ret = -1;
               break;
            }

            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, cal);  
         }
      }
      lFreeList(&direntries);
   }

   DRETURN(ret);
}

int sge_read_ckpt_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;
   const char *ckpt;

   DENTER(TOP_LAYER, "sge_read_ckpt_list_from_disk");
   
   if (*list == NULL) {
      *list = lCreateList("", CK_Type);
   } 

   direntries = sge_get_dirents(directory);
   if(direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINCKPTINTERFACEDEFINITIONS);
      for_each(direntry, direntries) {
         ckpt = lGetString(direntry, ST_name);

         if (ckpt[0] != '.') {
            if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_CKPT_S, ckpt);
               printf("\n");
            }
            ep = cull_read_in_ckpt(directory, ckpt, 1, 0, NULL, NULL);
            if (!ep) {
               DRETURN(-1);
            }

            if (ckpt_validate(ep, NULL)!=STATUS_OK) {
               lFreeElem(&ep);
               DRETURN(-1);
            }
            
            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, ckpt);
         }
      }
      lFreeList(&direntries);
   }
   DRETURN(0);
}

int sge_read_qinstance_list_from_disk(lListElem *cqueue)
{
   lList *dir_list;
   dstring qinstance_dir = DSTRING_INIT;
   const char *cqueue_name = lGetString(cqueue, CQ_name);

   DENTER(TOP_LAYER, "sge_read_qinstance_list_from_disk");

   sge_dstring_sprintf(&qinstance_dir, "%s/%s", QINSTANCES_DIR, cqueue_name);
   if (sge_is_directory(sge_dstring_get_string(&qinstance_dir))) {
      dir_list = sge_get_dirents(sge_dstring_get_string(&qinstance_dir));
      if (dir_list) {
         lListElem *dir;
         lList *qinstance_list = lGetList(cqueue, CQ_qinstances);
         
         for_each(dir, dir_list) {
            const char *hostname = lGetString(dir, ST_name);
            lListElem *qinstance = NULL;

            if (hostname[0] != '.') {
               qinstance = cull_read_in_qinstance(
                                        sge_dstring_get_string(&qinstance_dir), 
                                        hostname, 1, 0, NULL, NULL);
               if (qinstance == NULL) {
                  ERROR((SGE_EVENT, MSG_CONFIG_READINGFILE_SS, 
                         sge_dstring_get_string(&qinstance_dir), hostname));
                  DRETURN(-1);
               }
             
               if (qinstance_list == NULL) {
                  qinstance_list = lCreateList("", QU_Type);
                  lSetList(cqueue, CQ_qinstances, qinstance_list);
               } 
               lAppendElem(qinstance_list, qinstance);
            } else {
               sge_unlink(sge_dstring_get_string(&qinstance_dir), hostname);
            }
         }
         lFreeList(&dir_list);
      }
   }
   sge_dstring_free(&qinstance_dir);

   DRETURN(0);
}   

int sge_read_cqueue_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *qep, *direntry;
   int config_tag = 0;

   DENTER(TOP_LAYER, "sge_read_cqueue_list_from_disk");

   if (*list == NULL) {
      *list = lCreateList("", CQ_Type);
   }

   direntries = sge_get_dirents(directory);
   if (direntries) {
      const char *queue_str;
      
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINQUEUES);
      for_each(direntry, direntries) {

         queue_str = lGetString(direntry, ST_name);
         if (queue_str[0] != '.') {
            config_tag = 0;
            if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_QUEUE_S, queue_str);
               printf("\n");
            }
            if (verify_str_key(NULL, queue_str, MAX_VERIFY_STRING,
                               "cqueue", KEY_TABLE) != STATUS_OK) {
               DRETURN(-1);
            }   
            qep = cull_read_in_cqueue(directory, 
                                      queue_str, 1, 
                                      0, &config_tag, NULL);
            if (!qep) {
               ERROR((SGE_EVENT, MSG_CONFIG_READINGFILE_SS, directory, 
                      queue_str));
               DRETURN(-1);
            }
            
            if (!strcmp(lGetString(direntry, ST_name), SGE_TEMPLATE_NAME) && 
                !strcmp(lGetString(qep, CQ_name), SGE_TEMPLATE_NAME)) {
               /* 
                  we do not keep the queue template in the main queue list 
                  to be compatible with other old code in the qmaster
               */
               lFreeElem(&qep);
               sge_unlink(directory, lGetString(direntry, ST_name));
               WARNING((SGE_EVENT, MSG_CONFIG_OBSOLETEQUEUETEMPLATEFILEDELETED));
            }
            else if (!strcmp(lGetString(qep, CQ_name), SGE_TEMPLATE_NAME)) {
               /*
                  oops!  found queue 'template', but not in file 'template'
               */
               ERROR((SGE_EVENT, MSG_CONFIG_FOUNDQUEUETEMPLATEBUTNOTINFILETEMPLATEIGNORINGIT));
               lFreeElem(&qep);
            } else {
               lList *qinstance_list = NULL;
               lListElem *qinstance = NULL;

               sge_read_qinstance_list_from_disk(qep);
               qinstance_list = lGetList(qep, CQ_qinstances);
               for_each(qinstance, qinstance_list) {
                  lList *master_list = *(centry_list_get_master_list());
                  lList *ccl = NULL;

                  /* 
                   * handle slots from now on as a consumble 
                   * attribute of queue 
                   */
                  qinstance_set_conf_slots_used(qinstance);

                  /* setup actual list of queue */
                  qinstance_debit_consumable(qinstance, NULL, master_list, 0);

                  ccl = lGetList(qinstance, QU_consumable_config_list);
                  centry_list_fill_request(ccl, NULL, master_list, 
                                           true, false, true); 
                  if (ensure_attrib_available(NULL, qinstance, QU_load_thresholds) ||
                      ensure_attrib_available(NULL, qinstance, QU_suspend_thresholds) ||
                      ensure_attrib_available(NULL, qinstance, QU_consumable_config_list)) {
                     lFreeElem(&qep);
                     DRETURN(-1);
                  }
                  qinstance_state_set_unknown(qinstance, true);
                  qinstance_state_set_cal_disabled(qinstance, false);
                  qinstance_state_set_cal_suspended(qinstance, false);
                  qinstance_set_slots_used(qinstance, 0);
                  
                  if (host_list_locate(*object_type_get_master_list(SGE_TYPE_EXECHOST), 
                                       lGetHost(qinstance, QU_qhostname)) == NULL) {

                     ERROR((SGE_EVENT, MSG_CONFIG_CANTRECREATEQEUEUE_SS,
                            lGetString(qinstance, QU_qname), 
                            lGetHost(qinstance, QU_qhostname)));
                     lFreeElem(&qep);
                     lFreeList(&direntries);
                     DRETURN(-1);
                  } 
               }
               cqueue_list_add_cqueue(*list, qep);
            }
         } else {
            sge_unlink(directory, queue_str);
         }
      }
      lFreeList(&direntries);
   }
   

   DRETURN(0);
}

int sge_read_rqs_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *dir_entries = NULL;
   DENTER(TOP_LAYER, "sge_read_rqs_list_from_disk");

   if (*list == NULL) {
      *list = lCreateList("", RQS_Type);
   }

   dir_entries = sge_get_dirents(directory);
   if (dir_entries != NULL) {
      lListElem *dir_entry = NULL;
      
      if (!sge_silent_get()) {
         printf("%s\n", MSG_CONFIG_READINGINRQS);
      }

      for_each(dir_entry, dir_entries) {
         lListElem *rqs = NULL;
         dstring filename = DSTRING_INIT;
         const char *rqs_name;

         rqs_name = lGetString(dir_entry, ST_name);
         if (rqs_name[0] != '.') {
           lList *tmp_rqs_list = NULL;
           if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_RQS_S, rqs_name);
               printf("\n");
           }
           if (verify_str_key(NULL, rqs_name, MAX_VERIFY_STRING,
                              "rqs", KEY_TABLE) != STATUS_OK) {
               sge_dstring_free(&filename);
               lFreeList(&dir_entries);
               DRETURN(-1);
           }
           sge_dstring_sprintf(&filename, "%s/%s", directory, rqs_name);
           tmp_rqs_list = cull_read_in_rqs_list(sge_dstring_get_string(&filename), alpp);
           rqs = lCopyElem(lFirst(tmp_rqs_list));
           lFreeList(&tmp_rqs_list);

           if (!rqs) {
               ERROR((SGE_EVENT, MSG_CONFIG_READINGFILE_SS, directory, 
                      rqs_name));
               sge_dstring_free(&filename);
               lFreeList(&dir_entries);
               DRETURN(-1);
           }

           if (!rqs_verify_attributes(rqs, alpp, true)) {
               ERROR((SGE_EVENT, MSG_CONFIG_READINGFILE_SS, directory, 
                      rqs_name));
               lFreeElem(&rqs);
               sge_dstring_free(&filename);
               lFreeList(&dir_entries);
               DRETURN(-1);
           }

           lAppendElem(*list, rqs);

         } else {
            sge_unlink(directory, rqs_name);
         }
         sge_dstring_free(&filename);
      }
      lFreeList(&dir_entries);
   }

   DRETURN(0);
}

int sge_read_project_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;
   int config_tag = 0;

   DENTER(TOP_LAYER, "sge_read_project_list_from_disk");

   if (*list == NULL) {
      *list = lCreateList("", UP_Type);
   }
   
   direntries = sge_get_dirents(directory);
   if (direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINPROJECTS);

      for_each(direntry, direntries) {
         const char *userprj_str;

         userprj_str = lGetString(direntry, ST_name);
         if (userprj_str[0] != '.') {
            config_tag = 0;
            if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_PROJECT_S, lGetString(direntry, ST_name));
               printf("\n");
            }
            if (verify_str_key(NULL, userprj_str, MAX_VERIFY_STRING,
                               "project", KEY_TABLE) != STATUS_OK) {
               DRETURN(-1);
            }  
            ep = cull_read_in_userprj(directory, lGetString(direntry, ST_name), 1,
                                       0, &config_tag);
            if (!ep) {
               ERROR((SGE_EVENT, MSG_CONFIG_READINGFILE_SS, directory, 
                      lGetString(direntry, ST_name)));
               DRETURN(-1);
            }
            if (strcmp(lGetString(ep, UP_name), lGetString(direntry, ST_name))) {
               ERROR((SGE_EVENT, MSG_QMASTER_PRJINCORRECT_S,
                      lGetString(direntry, ST_name)));
               DRETURN(-1);
            }
            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, userprj_str);
         }
      }
      lFreeList(&direntries);
   }

   DRETURN(0);
}   

int sge_read_user_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;
   int config_tag = 0;

   DENTER(TOP_LAYER, "sge_read_user_list_from_disk");

   if (*list == NULL) {
      *list = lCreateList("", UP_Type);
   }

   direntries = sge_get_dirents(directory);
   if (direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINUSERS);
      
      for_each(direntry, direntries) {
         const char *direntry_str;
         
         direntry_str = lGetString(direntry, ST_name); 
         if (direntry_str[0] != '.') { 
            config_tag = 0;
            if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_USER_S, lGetString(direntry, ST_name));
               printf("\n");
            }

            ep = cull_read_in_userprj(directory, lGetString(direntry, ST_name), 1,
                                       1, &config_tag);
            if (!ep) {
               ERROR((SGE_EVENT, MSG_CONFIG_READINGFILE_SS, directory, 
                        lGetString(direntry, ST_name)));
               DRETURN(-1);
            }

            lAppendElem(*list, ep);
         } else {
            sge_unlink(directory, direntry_str);
         }
      }
      lFreeList(&direntries);
   }

   DRETURN(0);
}

int sge_read_userset_list_from_disk(lList **list, const char *directory, lList **alpp)
{
   lList *direntries;
   lListElem *ep, *direntry;

   DENTER(TOP_LAYER, "sge_read_userset_list_from_disk");

   if (*list == NULL) {
      *list = lCreateList("", US_Type);
   }

   direntries = sge_get_dirents(directory);
   if (direntries) {
      if (!sge_silent_get()) 
         printf("%s\n", MSG_CONFIG_READINGINUSERSETS);

      for_each(direntry, direntries) {
         const char *userset = lGetString(direntry, ST_name);

         if (userset[0] != '.') {
            if (!sge_silent_get()) {
               printf("\t");
               printf(MSG_SETUP_USERSET_S , lGetString(direntry, ST_name));
               printf("\n");
            }
            if (verify_str_key(NULL, userset, MAX_VERIFY_STRING,
                               "userset", KEY_TABLE) != STATUS_OK) {
               DRETURN(-1);
            }  

            ep = cull_read_in_userset(directory, userset, 1, 0, NULL); 
            if (!ep) {
               ERROR((SGE_EVENT, MSG_CONFIG_READINGFILE_SS, directory, 
                        userset));
               DRETURN(-1);
            }

            if(userset_validate_entries(ep, NULL, 1) == STATUS_OK) {
               lAppendElem(*list, ep);
            } else {
               lFreeElem(&ep);
            }
         } else {
            sge_unlink(directory, userset);
         }
      }
      lFreeList(&direntries);
   }
   
   DRETURN(0);
}


static int reresolve_host(
lListElem *ep,
int nm,
const char *object_name,
const char *object_dir 
) {
   char *old_name;
   const char *new_name;
   int ret;
   int pos;
   int dataType;

   DENTER(TOP_LAYER, "reresolve_host");

   pos = lGetPosViaElem(ep, nm, SGE_DO_ABORT);
   dataType = lGetPosType(lGetElemDescr(ep),pos);
   if (dataType == lHostT) {
      old_name = strdup(lGetHost(ep, nm));
   } else {
      old_name = strdup(lGetString(ep, nm));
   }
   ret = sge_resolve_host(ep, nm);
   if (ret != CL_RETVAL_OK ) {
      if (ret != CL_RETVAL_GETHOSTNAME_ERROR ) {
         /* finish qmaster setup only if hostname resolving
            does not work at all generally or a timeout
            indicates that commd itself blocks in resolving
            a host, e.g. when DNS times out */
         ERROR((SGE_EVENT, MSG_CONFIG_CANTRESOLVEHOSTNAMEX_SSS, object_name, old_name, cl_get_error_text(ret)));
         free(old_name);
         DRETURN(-1);
      }
      WARNING((SGE_EVENT, MSG_CONFIG_CANTRESOLVEHOSTNAMEX_SS, object_name, old_name));
   }

   /* rename config file if resolving changed name */
   if (dataType == lHostT) {
      new_name = lGetHost(ep, nm);
   } else {
      new_name = lGetString(ep, nm);
   }
   if (strcmp(old_name, new_name)) {
      if (!write_host(1, 2, ep, nm, NULL)) {
         ERROR((SGE_EVENT, MSG_SGETEXT_CANTSPOOL_SS, object_name, new_name));
         free(old_name);
         DRETURN(-1);
      }
      sge_unlink(object_dir, old_name);
   }
   free(old_name);

   DRETURN(0);
}

int read_all_centries(lList **list, const char *directory, lList **alpp)
{
   DIR *dir;
   SGE_STRUCT_DIRENT *dent;
   char dirent[SGE_PATH_MAX*2];
   char fstr[256];
   int fd;
   lListElem *el;
   lList *answer = NULL;


   DENTER(TOP_LAYER, "read_all_centries");

   if (*list == NULL) {
      *list = lCreateList("", CE_Type);
   }

   dir = opendir(directory);
   if (!dir) {
      ERROR((SGE_EVENT, MSG_FILE_NOOPENDIR_S, directory));
      DRETURN(-1);
   }
   if (!sge_silent_get())
      printf("%s\n", MSG_CONFIG_READINGINCOMPLEXATTRS);

   while (SGE_READDIR_R(dir, (SGE_STRUCT_DIRENT *)dirent, &dent)==0 && dent!=NULL) {
      if (!strcmp(dent->d_name,"..") || !strcmp(dent->d_name,".")) {
         continue;
      }
#if 0
      /* Enable this code for debugging */
      if (!sge_silent_get()) {
         printf("\t");
         printf(MSG_SETUP_COMPLEX_ATTR_S, dent->d_name);
         printf("\n");
      }  
#endif
      if ((dent->d_name[0] == '.')) {
         sge_unlink(directory, dent->d_name);
         continue;
      }

      sprintf(fstr, "%s/%s", directory, dent->d_name);
      
      if ((fd=SGE_OPEN2(fstr, O_RDONLY)) < 0) {
         ERROR((SGE_EVENT, MSG_FILE_NOOPEN_SS, fstr, strerror(errno)));
         continue;
      }
      close(fd);
      el = cull_read_in_centry(directory, dent->d_name , 1,0, *list);
      if (answer) {
         ERROR((SGE_EVENT, lGetString(lFirst(answer), AN_text)));
         lFreeList(&answer);
         DRETURN(-1);
      }
      if (el) {
         lAppendElem(*list, el);
      }
   }

   closedir(dir);

   centry_list_sort(*list);

   DRETURN(0);
}

/*----------------------------------------------------
 * read_all_configurations
 * qmaster function to read all configurations
 * should work with absolute pathnames
 * Must be called after internal setup
 *----------------------------------------------------*/

 
int read_all_configurations(lList **alpp,
                            lList **lpp, 
                            const char *global_config_file, 
                            const char *local_config_dir)
{
   DIR *dir;
   SGE_STRUCT_DIRENT *dent;
   char dirent[SGE_PATH_MAX*2];
   char fstr[256];
   lListElem *el;
   int ret;

   DENTER(TOP_LAYER, "read_all_configurations");

   if (!lpp) {
      DRETURN(-1);
   }

   if (local_config_dir == NULL) {
      DRETURN(-1);
   }

   if (!*lpp) {
      *lpp = lCreateList("conf list", CONF_Type);
   }

   /* First read global configuration. */
   el = read_configuration(global_config_file, SGE_GLOBAL_NAME, FLG_CONF_SPOOL);
   if (el)
      lAppendElem(*lpp, el);
   else {
      DRETURN(-1);
   }

   /* read local configurations from local_conf_dir */ 
   dir = opendir(local_config_dir);
   if (!dir) {
      DRETURN(-2);
   }

   while (SGE_READDIR_R(dir, (SGE_STRUCT_DIRENT *)dirent, &dent)==0 && dent!=NULL) {
      if (!dent->d_name)
                  continue;              /* May happen */
      if (!dent->d_name[0])
                  continue;              /* May happen */
                              
      if (!strcmp(dent->d_name,"..") || !strcmp(dent->d_name,"."))
         continue;

      sprintf(fstr, "%s/%s", local_config_dir, dent->d_name);
      
      el = read_configuration(fstr, dent->d_name, FLG_CONF_SPOOL);
      if (!el)
         continue;

      {
         char fname[SGE_PATH_MAX], real_fname[SGE_PATH_MAX];
         const char *new_name;
         char *old_name;
         lList *alp = NULL;

         /* resolve config name */
         old_name = strdup(lGetHost(el, CONF_hname));

         ret = sge_resolve_host(el, CONF_hname);
         if (ret != CL_RETVAL_OK) {
            if (ret != CL_RETVAL_GETHOSTNAME_ERROR  ) {
               ERROR((SGE_EVENT, MSG_CONFIG_CANTRESOLVEHOSTNAMEX_SSS, "local configuration", old_name, cl_get_error_text(ret)));
               free(old_name);
               DRETURN(-1);
            }
            WARNING((SGE_EVENT, MSG_CONFIG_CANTRESOLVEHOSTNAMEX_SS, "local configuration", old_name));
         }
         new_name = lGetHost(el, CONF_hname);

         /* simply ignore it if it exists already */
         if (*lpp && lGetElemHost(*lpp, CONF_hname, new_name)) {
            free(old_name);
            lFreeElem(&el);
            continue;
         }

         /* rename config file if resolving changed name */
         if (strcmp(old_name, new_name)) {

            sprintf(fname, "%s/.%s", local_config_dir, new_name);
            sprintf(real_fname, "%s/%s", local_config_dir, new_name);

            DPRINTF(("global_config_file: %s\n", fname));
            if ((ret=write_configuration(1, &alp, fname, el, NULL, FLG_CONF_SPOOL))) {
               /* answer list gets filled in write_configuration() */
               free(old_name);
               sge_switch2start_user();
               lFreeList(&alp);
               DRETURN(-1);
            } else {
               char old_fname[SGE_PATH_MAX];

               if (rename(fname, real_fname) == -1) {
                  free(old_name);
                  DRETURN(-1);
               }
               sprintf(old_fname, "%s/%s", local_config_dir, old_name);
               if (!sge_unlink(NULL, old_fname)) {
                  DRETURN(-1);
               }
            }
         }
         lFreeList(&alp);
         FREE(old_name);
      }

      lAppendElem(*lpp, el);
   }

   closedir(dir);

   DRETURN(0);
}


